# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

row_mean_dgcmatrix <- function(x, i, rows, cols) {
    .Call('_sctransform_row_mean_dgcmatrix', PACKAGE = 'sctransform', x, i, rows, cols)
}

row_gmean_dgcmatrix <- function(x, i, rows, cols, eps) {
    .Call('_sctransform_row_gmean_dgcmatrix', PACKAGE = 'sctransform', x, i, rows, cols, eps)
}

row_var_dgcmatrix <- function(x, i, rows, cols) {
    .Call('_sctransform_row_var_dgcmatrix', PACKAGE = 'sctransform', x, i, rows, cols)
}

row_var_dense_d <- function(x) {
    .Call('_sctransform_row_var_dense_d', PACKAGE = 'sctransform', x)
}

row_var_dense_i <- function(x) {
    .Call('_sctransform_row_var_dense_i', PACKAGE = 'sctransform', x)
}

